/* User Interface - Menu
 * (c) by Marek Zelem
 */

#ifndef _MENU_H
#define	_MENU_H

#include <stdio.h>
#include <ctype.h>

struct typehandler_t	{
	long (*input)( int fd );
	void (*free)( long d );
	long (*dup)( long d );
	/* store: zapise d na addr, vrati pocet zabratych bajtov ak *(-1), */
	/*	  tak je to odporucany koncovy prvok pola	*/
	long (*store)( void *addr, long d );
	long (*read)( void *addr );
	long (*print)( int fd, long d );	/* vracia ako store */
			};

typedef struct	{
	char			*name;
	struct typehandler_t	*handler;
		} types_t;

extern struct typehandler_t
			th_s_char,th_s_short,th_s_int,th_s_long,
			th_u_char,th_u_short,th_u_int,th_u_long,
			th_s_char_p,th_u_char_p,
			th_void,th_void_p;

#define	END_TYPES	{NULL,NULL}
#define STD_TYPES	{"char",&th_s_char},\
			{"char *",&th_s_char_p},\
			{"signed char *",&th_s_char_p},\
			{"unsigned char *",&th_u_char_p},\
			{"short",&th_s_short},\
			{"short int",&th_s_short},\
			{"int",&th_s_int},\
			{"long",&th_s_long},\
			{"long int",&th_s_long},\
			{"signed char",&th_s_char},\
			{"signed short",&th_s_short},\
			{"signed short int",&th_s_short},\
			{"signed int",&th_s_int},\
			{"signed",&th_s_int},\
			{"signed long",&th_s_long},\
			{"signed long int",&th_s_long},\
			{"unsigned char",&th_u_char},\
			{"unsigned short",&th_u_short},\
			{"unsigned short int",&th_u_short},\
			{"unsigned int",&th_u_int},\
			{"unsigned",&th_u_int},\
			{"unsigned long",&th_u_long},\
			{"unsigned long int",&th_u_long},\
			{"s_char",&th_s_char},\
			{"s_short",&th_s_short},\
			{"s_int",&th_s_int},\
			{"s_long",&th_s_long},\
			{"u_char",&th_u_char},\
			{"u_short",&th_u_short},\
			{"u_int",&th_u_int},\
			{"u_long",&th_u_long},\
			{"void",&th_void},\
			{"void *",&th_void_p}

int declcmp( const char *a, const char *b );

/* name je deklaracia premennej aj s menom.
 * priklad: "unsigned int pole[3]"
 * 	    "char pole[]" == "char *pole"
 * 	    "char pole[3][4]"
 * 	    "int i"
 */
char *type_name( const char *name );
char *type_varname( const char *name );
int *type_size( const char *name );

long type_input( int fd, const char *name, types_t *types );
int type_print( int fd, const char *name, types_t *types, long d );
int type_free( const char *name, types_t *types, long d );
long type_dup( const char *name, types_t *types, long d );

long type_str_input( char *buf, const char *name, types_t *types );

char *read_string( int fd );

typedef struct	{
	char	*type;		/* ak type ==NULL ,tak koniec */
	int	offset;
	int	len;		/* 1-4 bytes ; -1 - pointer sem */
		} struct_t;

/* asi zle
#define STRUCT_MEMBER(stru,mem)	\
	{#mem,(int)(&((((stru) *)0)->(mem))),sizeof((((stru) *)0)->(mem)) }
*/

struct 	vars	{
	struct vars	*next;
//	char	*name;
	char	*decl;
	long	value;
		};
typedef struct	{
	types_t	*types;
	struct vars	*vars;
		} vars_t;

extern int val_err;
vars_t *var_init( types_t *types );
int var_set( vars_t *vars, char *decl, long value );
long var_getval( vars_t *vars, char *decl );
int var_del( vars_t *vars, char *decl );
int var_stop( vars_t *vars );

#define	MENU_MASK	0xff00
#define	MENU_NOP	0x0000
#define	MENU_INIT	0x0100		/* pri inicializacii */
#define	MENU_FINISH	0x0200		/* pri ukonceni menu */
#define	MENU_EXIT	0x0300		/* ako ITEM ale potom menu skonci */
#define	MENU_ITEM	0x0400		/* pri zadani prikazu */
#define	MENU_SUB	0x0500		/* pri zadani prikazu - submenu */
#define	MENU_END	0xffff

typedef struct	{
	int	typ;
	char	*command;
	char	*help;
	char	*retval_dec;
	void	*func;
	char	*param_dec;
		} menu_t;

typedef struct	s_MENU_t {
	struct s_MENU_t	*prev;
	menu_t	*menu;
	int	fd;
	vars_t	*vars;
	char	*retvars;
	char	*prompt;
	char	*not_found;
		} MENU_t;

long func_call( void *func, long *par, long n );	/* asm */

int menu_exec_function( int fd, vars_t *vars, vars_t *local, menu_t *menu );
MENU_t *menu_init( int fd, types_t *types, menu_t *menu, vars_t *vars );
MENU_t *menu_exit( MENU_t *m );
int menu_setvar( MENU_t *m, char *decl, long value );
int menu_do_command( MENU_t *m, int typ, char *cmd );
MENU_t *menu_enter_sub( MENU_t *m, char *cmd );

/* menu_text_command - uzavrie MENU_t *m */
int menu_text_command( MENU_t *m );

#endif	_MENU_H

