/* process's Modules library
 * (c) by Marek Zelem
 */

#ifndef _MODULES_H
#define _MODULES_H

#ifdef linux
#include <linux/types.h>
#else
#include <machine/types.h>
#endif
#include <mmm/elf.h>

#define	E_UNDEF		-200
#define	E_NOINIT	-201		/* neexisuje init_module */
#define	E_NOMOD		-210
#define	E_MODEXIST	-211
#define	E_MODUNINIT	-212		/* uninitialized module */
#define	E_MODSYMUSED	-213		/* mod->used!=NULL */
#define	E_MODINUSE	-214		/* mod->usecount!=0 */

#define	E_IMPOSIBLE	-500		/* nemoze vzniknut ! (nemalo by) */

typedef struct	{
	char	*name;
	void	*addr;
		} mod_syms_t;
#define SF(x)	{#x,(x)}
#define	SD(x)	{#x,&(x)}
#ifndef SE
#define	SE	{NULL,NULL}
#endif	SE

struct s_modsym	{
	char	*name;						/* none */
	long	value;
	u_long	size;
	u_short	segment;
	u_short	type;	/* unused */
		};

struct s_modrel	{
	long	*addr;
	long	addend;
	int	symbol;
	int	type;
		};
		
struct s_modseg	{
	char	*name;						/* free */
	char	*base;
	elfmap_t map;						/* unmap */
	long	size;
	int	prot;
	int	section;
		};
		
struct s_modules_t;

struct s_moduse	{
	struct s_moduse	*next_used;
	struct s_moduse	*prev_used;
	struct s_moduse	*next_use;
	struct s_moduse	*prev_use;
	struct s_modules_t	*module_use;
	struct s_modules_t	*module_used;
	int	count;
		};

/* status */
#define	MS_INITIALIZED	0x01
#define MS_SET_MASK	0x06
#define	MS_AUTOINIT	0x02
#define	MS_AUTOCLEAN	0x04
		
typedef struct s_modules_t	{
	struct s_modules_t	*next;
	struct s_modules_t	*prev;
	struct s_moduse	*use;		/* pouziva */
	struct s_moduse *used;		/* je pouzivany */
	int		usecount;
	int		status;
	char		*name;					/* free */
	char		*symnames;
	elfmap_t	map_symnames;			/* unmap */
	struct s_modsym	*sym;					/* free */
	u_long		nr_sym,sym_size;
	struct s_modseg	*segment;				/* free */
	u_long		nr_seg;
	struct s_modsym	*undef;					/* free */
	long	undef_len,undef_size;
	struct s_modrel	*rel;					/* free */
	long	rel_len,rel_size;
				} modules_t;

int modules_verlib( char *buf );

modules_t *modules_start( mod_syms_t *syms );
int modules_stop( modules_t *mod );
int modules_add_syms( modules_t *mod, mod_syms_t *syms );

struct s_modsym *modules_get_symbol( modules_t *me, modules_t *mod,
					char *name );
int modules_put_symbol( modules_t *me, modules_t *mod, struct s_modsym *sym );

/* Ak existuje funkcia int __loaded_module(void),
 * tak ju zavola, ak vrati <0 tak error
 * Ak funkcia neexistuje a existuju nedefinovane pouzite symboly, tak err
 */
int modules_load_module( modules_t *mod, const char *filename,
		const char *name, int status );

modules_t *modules_find_module( modules_t *mod, const char *name );

/* Ak m->rel!=NULL (existuju nedefinovane pouzite symboly )
 *	tak ak existuje funkcia int __test_module(void) a vrati >=0
 *	    tak pokracuj v inicializacii
 *	    inak error;
 * Ak existuje funkcia int init_module(void) a vrati >=0
 *	tak ok;
 *	inak Err;
 */
int modules_init_module( modules_t *mod, const char *name );

/* Ak m->used!=NULL (symboly modulu pouziva iny modul) tak error
 * Ak usecount!=0, tak ak existuje funkcia int __finish_module(void)
 *		tak ju zavola a pokracuje.
 * Ak usecount!=0, tak error
 * Odpoji m->next a m->prev
 * Ak je modul inicializovany a existuje funkcia int cleanup_module(void)
 *		tak ju zavola.
 */
int modules_remove_module( modules_t *mod, const char *name );

int modules_print_undef( int fd, modules_t *mod );
int modules_print( int fd, modules_t *mod, int level );
/* level: 0-iba moduli, 1-used by, 2-use, 3-symbols, 4-undef syms */

void modules_inc_usecount( modules_t *mod );
void modules_dec_usecount( modules_t *mod );

/* symboly definovane pre moduly */
extern int modules_lib_version;
extern char modules_lib_strversion[];
extern modules_t __this_module[];
extern modules_t __modules_head[];
#define mod_INC_USE_COUNT	modules_inc_usecount(__this_module)
#define mod_DEC_USE_COUNT	modules_dec_usecount(__this_module)

#endif	_MODULES_H

