/* ionet_lib.c	  */
/* (c) by Marek Zelem */

#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include "io_lib.h"
#include "dynamic.h"

#define	MINIMAL_LENGTH_OF_BUFFER	128

/* iof_in_out
 * pre RD: cita z rd a zapisuje do inbuf a vola ufn()
 * pre WR: cita z outbuf a zapisuje do wr, ak je prazdny outbuf, tak vola ufn()
 * in/out buffer je typu dfifo_t
 * na vsetky ostatne akcie vola ufn()
 * ak rd,wr a ex zatvorene, tak generuje akciu IO_CLOSED
 * ak sa zapis do wr nepodari, tak sa objekt ukonci
 */
int iof_in_out( io_t *io, int what, void *arg )
{ int l,lw;
  char buf[MINIMAL_LENGTH_OF_BUFFER];
	if( what==IO_RD )
	{	if( (l=read(io->rd,buf,MINIMAL_LENGTH_OF_BUFFER))>0 )
		{	if( dfifo_write(io->inbuf,buf,l)!=l )
				return(-1);
			return( io->ufn(io,what,arg) );
		}
		else if( l==0 )
		{	io_chrd(io,-1);
			if( io->outbuf==NULL )	return(-1);
		}
		else return( io->ufn(io,IO_RDERR,arg) );
	}
	else if( what==IO_WR )
	{	if( io->outbuf==NOTNULL )
			return( io->ufn(io,IO_WREXTR,arg) );
		if( (l=dfifo_read(io->outbuf,buf,MINIMAL_LENGTH_OF_BUFFER))>0 )
		{	if( (lw=write(io->wr,buf,l))!=l )
			{	if( lw<0 )	return(-1);
				if( (l-lw)>0)
				  if( dfifo_unread(io->outbuf,buf+lw,l-lw)
					  	!=(l-lw) )
					return(-1);
			}
		}
		else if( l==0 )	return( io->ufn(io,what,arg) );
		else return( io->ufn(io,IO_WRERR,arg) );
	}
	else if( what==IO_CLOSED )	return( -1 );
//	else if( what==IO_CLOSED )	return( io->ufn(io,what,arg) );
	else if( what==IO_INIT )
	{   if( io->ufn==NULL || io->ufn(io,what,arg)<0 )	return(-1);
	    if( io->rd>=0 && io->inbuf==NULL )
		if((io->inbuf=dfifo_create(1,MINIMAL_LENGTH_OF_BUFFER))==NULL)
			return(-1);
	}
	else if( what==IO_DEL )
	{	if( io->ufn!=NULL )	io->ufn(io,what,arg);
		if( io->inbuf!=NULL && io->inbuf!=NOTNULL )
			dfifo_delete(io->inbuf);
		if( io->outbuf!=NULL && io->outbuf!=NOTNULL )
			dfifo_delete(io->outbuf);
	}
	else	return( io->ufn(io,what,arg) );
	return(0);
}

/* iof_input_line
 * pre RD: cita z rd a zapisuje do inbuf, ak \n tak vola ufn()
 * inak ako pre iof_in_out
 */
int iof_input_line( io_t *io, int what, void *arg )
{ char buf[MINIMAL_LENGTH_OF_BUFFER];
  int r,l,i;
	if( what==IO_RD )
	{	if( (l=read(io->rd,buf,MINIMAL_LENGTH_OF_BUFFER))>0 )
		{	i=0;
			while( i<l )
			{	if( dfifo_write(io->inbuf,buf+i,1)!=1 )
					return(-1);
				if( buf[i]=='\n' )
				{	if( (r=io->ufn(io,what,arg))<0 )
						return(r);
				}
				i++;
			}
		}
		else if( l==0 )
		{	io_chrd(io,-1);
			if( io->outbuf==NULL )	return(-1);
		}
		else return( io->ufn(io,IO_RDERR,arg) );
	}
	else	return(iof_in_out(io,what,arg));
	return(0);
}

/* iof_input_char
 * pre RD: cita z rd a zapisuje do inbuf, za kazdym znakom vola ufn()
 * inak ako pre iof_in_out
 */
int iof_input_char( io_t *io, int what, void *arg )
{ char buf[8];
  int r,l,i;
	if( what==IO_RD )
	{	if( (l=read(io->rd,buf,8))>0 )
		{	i=0;
			while( i<l )
			{	if( dfifo_write(io->inbuf,buf+i,1)!=1 )
					return(-1);
				if( (r=io->ufn(io,what,arg))<0 )
					return(r);
				i++;
			}
		}
		else if( l==0 )
		{	io_chrd(io,-1);
			if( io->outbuf==NULL )	return(-1);
		}
		else return( io->ufn(io,IO_RDERR,arg) );
	}
	else	return(iof_in_out(io,what,arg));
	return(0);
}

/* ---- funkcie ---- */

int io_write( io_t *io, char *buf, int len )
{
	if( io->wr>=0 )
	{	if( io->outbuf == NULL )
			if( (io->outbuf=dfifo_create(1,128))==NULL )
				return(0);
			return(dfifo_write(io->outbuf,buf,len));
	}
	return(-1);
}

int io_bufwrite( io_t *io, dfifo_t *buf )
{
	if( io->wr>=0 )
	{	if( io->outbuf == NULL )
			if( (io->outbuf=dfifo_create(1,128))==NULL )
				return(0);
			return(dfifo_copy(buf,io->outbuf));
	}
	return(-1);
}

int io_printf( io_t *io, char *fmt, ... )
{ va_list ap;
  int max,len;
  char *buf;

	max=1024;
G_io_p:
  	if( (buf=malloc(max))==NULL )	return(-1);
  	va_start(ap,fmt);
	len=vsnprintf(buf,max,fmt,ap);
	if( len>=max )
	{	free(buf);
		max+=4096;	goto G_io_p;
	}
	va_end(ap);
	len=io_write(io,buf,len);
	free(buf);
	return(len);
}

int io_writebuf_erase( io_t *io )	/* return -1 if rd < 0 */
{	dfifo_delete(io->outbuf);
	io->outbuf=NULL;
	if( io->rd<0 ) return(-1);
	return(0);
}

