/* Memory in file
 * (c)24.10.1999 by Marek Zelem
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <mmm/fmemory.h>
#include <mmm/mutex.h>

#define NEXT(b)	((struct fmem_blok_head_s *)(((char*)(b))+(b)->size))
#define PREV(b)	((struct fmem_blok_head_s *)(((char*)(b))-(b)->prev))
#define NEW(b,size)	((struct fmem_blok_head_s *)(((char*)(b))+(size)))
#define BHEAD(p)	((struct fmem_blok_head_s *)(((char*)(p))-sizeof(struct fmem_blok_head_s)))

int	fmem_errno=0;

static int get_access( struct fmem_head_s *f )
{ int t=0;
	mutex_wait(&(f->mutex[0]))
	{	if( (t++)>=100 )
		{	fmem_errno=FME_MUTEX;
			return(0);
		}
		sched_yield();
	}
	fmem_errno=0;
	return(1);
}
static void put_access( struct fmem_head_s *f )
{
	mutex_signal(&(f->mutex[0]));
}

static void mem_destroy(struct fmem_s *m)
{ int i;
	for(i=0;i<m->files;i++)
	{	munmap(m->file[i].mem,m->file[i].size);
		if( m->file[i].fd>=0 )
			close(m->file[i].fd);
		if( m->file[i].filename!=NULL )
			free(m->file[i].filename);
	}
	free(m);
}

static int mem_add_file(struct fmem_s *m, char *file)
{ struct fmem_head_s *a;
  int fd;
  struct stat st;
	if( m->files >= m->max_files )
		return(-1);
	if( (fd=open(file,O_RDWR))<0 )
		return(-1);
	if( fstat(fd,&st)<0 )
	{	close(fd); return(-1);	}
	a=mmap(NULL,st.st_size,PROT_READ|PROT_WRITE,MAP_SHARED|MAP_FILE,fd,0);
	if( a==NULL || a==(void*)(-1) /*MAP_FAILED*/ )
	{	close(fd); return(-1);	}
	if( a->magic!=0x1ef1f11e ||
		m->fileid_trans_table[(a->file_magic_id)>>24]!=0xff )
	{	munmap(a,st.st_size); close(fd);
		return(-1);
	}
	m->file[m->files].filename=strdup(file);
	m->file[m->files].fd=fd;
	m->file[m->files].size=st.st_size;
	m->file[m->files].mem=a;
	m->fileid_trans_table[(a->file_magic_id)>>24]=m->files;
	m->files++;
	return(0);
}

static int write_null( int fd, int len )
{ char buf[10240];
	memset(buf,0,10240);
	while( len>10240 )
	{	if( write(fd,buf,10240)!=10240 )
			return(-1);
		len-=10240;
	}
	if( len>0 )
		if( write(fd,buf,len)!=len )
			return(-1);
	return(0);
}

static int mem_create_file( struct fmem_s *m, char *file, int id, long size )
{ struct fmem_head_s head;
  struct fmem_blok_head_s blok;
  int fd;
	if( (fd=open(file,O_WRONLY|O_CREAT|O_EXCL,0666))<0 )
		return(-1);
	size=ALIGN16(size);
	head.magic=0x1ef1f11e;
	head.file_magic_id=id<<24;
	head.first_offset=ALIGN16(sizeof(head));
	head.free=size-head.first_offset-sizeof(blok);
	head.mutex[0]=head.mutex[1]=head.mutex[2]=head.mutex[3]=0;
	blok.size=head.free;
	blok.prev=0;
	blok.usecount=0;
	blok.user_id=0;
	blok.user_adr=0;
	if( write(fd,&head,sizeof(head))!=sizeof(head) )
		goto Err;
	if( sizeof(head)<ALIGN16(sizeof(head)) )
		if( write_null(fd,ALIGN16(sizeof(head))-sizeof(head))<0 )
			goto Err;
	if( write(fd,&blok,sizeof(blok))!=sizeof(blok) )
		goto Err;
	if( write_null(fd,blok.size-sizeof(blok))<0 )
		goto Err;
	blok.prev=blok.size;
	blok.size=0;
	blok.usecount=0xffff;
	blok.user_id=0;
	blok.user_adr=0;
	if( write(fd,&blok,sizeof(blok))!=sizeof(blok) )
		goto Err;
	close(fd);
	return(0);
Err:	close(fd);
	return(-1);
}

static struct fmem_file_s *get_file( struct fmem_s *m, long addr )
{ int i;
	i=m->fileid_trans_table[addr>>24];
	if( i<0 || i>m->files )
		return(NULL);
	return( &(m->file[i]) );
}

static struct fmem_blok_head_s *get_block( struct fmem_s *m, long addr )
{ int i;
  struct fmem_blok_head_s *b;
	i=m->fileid_trans_table[addr>>24];
	if( i<0 || i>m->files )
		return(NULL);
	if( !get_access(m->file[i].mem) )	return(NULL);
	if( (addr&FMEMADDR_BLOKID)*16 < m->file[i].mem->first_offset ||
	    (addr&FMEMADDR_BLOKID)*16 >= m->file[i].size )
		goto End;
	b=(struct fmem_blok_head_s *)
		(((char*)(m->file[i].mem))+(addr&FMEMADDR_BLOKID)*16);
	if( !(b->usecount) )
		goto End;
	/* !!! este otestovat ci NEXT(b) nieje uplne mimo */
	if( NEXT(b)->prev!=b->size )
		goto End;
	put_access(m->file[i].mem);
	return( b );
End:	put_access(m->file[i].mem);
	return(NULL);
}

static void *mem_get( struct fmem_s *m, long addr )
{ struct fmem_blok_head_s *b;
	if( (b=get_block(m,addr))==NULL )
		return(NULL);
	return(b+1);
}

static long mem_get_size( struct fmem_s *m, long addr )
{ struct fmem_blok_head_s *b;
	if( (b=get_block(m,addr))==NULL )
		return(0);
	return( b->size - sizeof(struct fmem_blok_head_s) );
}

static void block_split(struct fmem_blok_head_s *b, long size )
{ struct fmem_blok_head_s *n;
	if( b->size<=size )	return;
	n= NEW(b,size);
	n->size=b->size-size;
	n->prev=size;
	n->usecount=0;
	n->user_id=0;
	n->user_adr=0;
	NEXT(n)->prev=n->size;
	b->size=size;
}
static struct fmem_blok_head_s *block_alloc( struct fmem_file_s *f, long size )
{ struct fmem_blok_head_s *b;
	if( f->mem->free<size )	return(NULL);
	size+=sizeof(struct fmem_blok_head_s);
	size=ALIGN16(size);
	b= (struct fmem_blok_head_s *)
		(((char*)(f->mem)) + f->mem->first_offset);
	while( b->size>0 )
	{	if( !(b->usecount) && b->size>=size )
		{	block_split(b,size);
			b->usecount=1;
			f->mem->free-=size;
			return(b);
		}
		b= NEXT(b);
	}
	return(NULL);
}

static long mem_malloc( struct fmem_s *m, long size, int memid )
{ struct fmem_blok_head_s *b;
  struct fmem_file_s *f;
	memid&=0x00ff;
	while( memid<255 && m->fileid_trans_table[memid]!=0xff )
	{	if( !get_access(m->file[m->fileid_trans_table[memid]].mem) )
			return(NOMEMID);
		b=block_alloc((f=&(m->file[m->fileid_trans_table[memid]])),
			size);
		put_access(m->file[m->fileid_trans_table[memid]].mem);
		if( b!=NULL )
		{	return(
			  (memid<<24)|((((char*)b)-((char*)(f->mem)))/16) );
		}
		memid++;
	}
	return(NOMEMID);
}

static void block_free( struct fmem_file_s *fi, struct fmem_blok_head_s *b )
{ struct fmem_blok_head_s *p,*f;
	fi->mem->free+=b->size;
	p=f=b;
	while( p->prev>0 )
	{	p= PREV(p);
		if( p->usecount )
			break;
		f=p;
	}
	while( b->size>0 )
	{	b= NEXT(b);
		if( b->usecount )
			break;
	}
	b->prev= ((char*)b)-((char*)f);
	f->size= ((char*)b)-((char*)f);
	f->usecount=0;
}

static int mem_free( struct fmem_s *m, long addr )
{ struct fmem_blok_head_s *b;
  struct fmem_file_s *fi;
  	fi=get_file(m,addr);
	if( (b=get_block(m,addr))==NULL )
		return(0);
	if( !get_access(fi->mem) )	return(-1);
	block_free(fi,b);
	put_access(fi->mem);
	return(0);
}

static int block_resize(struct fmem_s *m, long addr, long size)
{ struct fmem_blok_head_s *b,*n,*v;
  struct fmem_file_s *fi;
	fi=get_file(m,addr);
	if( (b=get_block(m,addr))==NULL )
		return(0);
	if( !get_access(fi->mem) )	return(-1);
	size+=sizeof(struct fmem_blok_head_s);
	size=ALIGN16(size);
	if( size == b->size )	return(1);
	if( size < b->size )
	{	block_split(b,size);
		block_free(fi,NEXT(b));
		put_access(fi->mem);
		return(1);
	}
	n= NEXT(b);
	if( !(n->usecount) && b->size+n->size>=size )
	{	fi->mem->free-=size - b->size;
		v= NEW(b,size);
		if( b->size+n->size > size )
		{	v->size=b->size+n->size-size;
			v->usecount=0;
			v->user_id=0;
			v->user_adr=0;
		}
		v->prev=size;
		b->size=size;
		put_access(fi->mem);
		return(1);
	}
	put_access(fi->mem);
	return(0);
}

static long mem_realloc( struct fmem_s *m, long addr, long size )
{ long new;
  char *f,*t;

	if( block_resize(m,addr,size) )
		return(addr);
	if( (f=mem_get(m,addr))==NULL )
		return(NOMEMID);
	new=mem_malloc(m,size,addr>>24);
	if( new==NOMEMID || (t=mem_get(m,new))==NULL )
		return(NOMEMID);
	if( !get_access(get_file(m,addr)->mem) )	return(-1);
	BHEAD(t)->usecount=BHEAD(f)->usecount;
	BHEAD(t)->user_id=BHEAD(f)->user_id;
	BHEAD(t)->user_adr=BHEAD(f)->user_adr;
	if( mem_get_size(m,addr)<size )
		memcpy(t,f,mem_get_size(m,addr));
	else	memcpy(t,f,size);
	put_access(get_file(m,addr)->mem);
	mem_free(m,addr);
	return(new);
}

static struct fmem_head_s *mem_file_head(struct fmem_s *m,int memid)
{ int i;
	memid&=0xff;
	i=m->fileid_trans_table[memid];
	if( i<0 || i>m->files )
		return(NULL);
	return( m->file[i].mem );
}

fmem_t *fmem_create( int max_files )
{ fmem_t *m;
  int i;
	if( sizeof(struct fmem_blok_head_s)!=16 )
		return(NULL);
	if( max_files<=0 || max_files>255 )
		return(NULL);
	if( (m=malloc(sizeof(fmem_t)+max_files*sizeof(struct fmem_file_s)))
			==NULL )
		return(NULL);
	for(i=0;i<256;i++)	m->fileid_trans_table[i]=0xff;
	m->max_files=max_files;
	m->files=0;
	m->destroy=mem_destroy;
	m->add_file=mem_add_file;
	m->create_file=mem_create_file;
	m->get=mem_get;
	m->get_size=mem_get_size;
	m->malloc=mem_malloc;
	m->realloc=mem_realloc;
	m->free=mem_free;
	m->head=get_block;
	m->file_head=mem_file_head;
	return(m);
}

